#![feature(box_syntax)]

static LOWEST_PRECEDENCE: usize = 0;

#[derive(Debug, Copy, Clone)]
enum Token<'a> {
	IntegerLiteral(&'a str),
	Plus,
	Semicolon,
	LeftParen,
	RightParen,
}

impl Token<'_> {
	// Returns length of that token in bytes, which is used for advancing the
	// cursor in the lexer.
	pub fn len(&self) -> usize {
		return match self {
			Token::IntegerLiteral(s) => s.len(),
			Token::Plus | Token::Semicolon | Token::LeftParen | Token::RightParen => 1,
		};
	}

	pub fn precedence(&self) -> usize {
		return match self {
			Token::Plus => 1,
			_ => LOWEST_PRECEDENCE,
		}
	}
}

struct Source<'a> {
	source: &'a str,
	cursor: usize,
	last: Option<Token<'a>>,
}

impl<'a> Source<'a> {
	pub fn new(source: &'a str) -> Self {
		return Self {
			source,
			cursor: 0,
			last: None,
		};
	}

	fn skip_whitespace(&mut self) {
		let mut chars = self.source[self.cursor..].chars();

		while let Some(c) = chars.next() {
			if c.is_whitespace() {
					self.cursor += c.len_utf8();
			} else {
				return;
			}
		}
	}

	fn get_next(&mut self) -> Option<Token<'a>> {
		self.skip_whitespace();
		let mut chars = self.source[self.cursor..].chars();

		let token = match chars.next()? {
			'+' => Token::Plus,
			';' => Token::Semicolon,
			'(' => Token::LeftParen,
			')' => Token::RightParen,
			c if c.is_ascii_digit() => {
				let start = self.cursor;
				let mut length = c.len_utf8();

				loop {
					match chars.next()? {
						c if c.is_ascii_digit() => length += c.len_utf8(),
						_ => break,
					};
				};

				Token::IntegerLiteral(&self.source[start..start + length])
			},
			c => todo!("invalid character `{:?}`", c)
		};

		return Some(token);
	}

	pub fn next(&mut self) -> Option<Token<'a>> {
		let token = match self.last {
			Some(t) => {
				self.last = None;
				t
			},
			None => self.get_next()?,
		};

		self.cursor += token.len();
		return Some(token);
	}

	pub fn peek(&mut self) -> Option<Token<'a>> {
		// We unwrap and then wrap it again as it makes more semantic sense, since
		// an Option that get_next returns is not connected to what peek returns.
		// In future we might want to add more sophisticated error handling to that
		// function, and then it will get easier to refactor. More so, we avoid
		// exposing lexer's internal state to the user.
		self.last = Some(self.get_next()?);
		return self.last;
	}
}

// Represents a dynamic parsing process, will get converted to ast::Tree after
// it completes.
struct Parser<'a> {
	source: &'a mut Source<'a>,
}

#[derive(Debug)]
enum Expr<'a> {
	Literal(&'a str),
	Paren(Box<Expr<'a>>),
	Binary(Token<'a>, Box<Expr<'a>>, Box<Expr<'a>>),
}

impl<'a> Parser<'a> {
	pub fn new(source: &'a mut Source<'a>) -> Self {
		return Self {
			source,
		};
	}

	#[inline(always)]
	fn next(&mut self) -> Option<Token<'a>> {
		return self.source.next();
	}

	#[inline(always)]
	fn peek(&mut self) -> Option<Token<'a>> {
		return self.source.peek();
	}

	#[inline(always)]
	fn bump(&mut self) {
		self.next();
	}

	fn parse_unary_expr(&mut self) -> Option<Expr<'a>> {
		return match self.next()? {
			Token::IntegerLiteral(s) => Some(Expr::Literal(s)),
			Token::LeftParen => {
				let expr = self.parse_expr(0)?;

				return match self.next()? {
					Token::RightParen => Some(Expr::Paren(box expr)),
					_ => None,
				};
			},
			_ => None,
		};
	}

	pub fn parse_expr(&mut self, min_precedence: usize) -> Option<Expr<'a>> {
		let mut lhs = self.parse_unary_expr()?;

		loop {
			match self.peek()? {
				token @ Token::Plus => {
					let prec = token.precedence();
					if prec <= min_precedence {
						return Some(lhs);
					};

					// Don't advance the parser before we make sure that the precedence
					// is correct.
					self.bump();

					let rhs = self.parse_expr(prec)?;
					lhs = Expr::Binary(token, box lhs, box rhs);
				},
				_ => return Some(lhs),
			};
		}
	}
}

fn main() {
	// let inline_source = "3 + 5 + 7;";
	// let inline_source = "(3 + 5) + 7;";
	let inline_source = "3 + (5 + (7 + 11));";
	let mut source = Source::new(inline_source);
	let mut parser = Parser::new(&mut source);
	eprintln!("{:?}", parser.parse_expr(0));
}
